/* ipv4/icmp.c
 *
 * rtnet - real-time networking subsystem
 * Copyright (C) 1999,2000 Zentropic Computing, LLC
 *               2002 Ulrich Marx <marx@kammer.uni-hannover.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <linux/spinlock.h>
#include <linux/ip.h>
#include <linux/icmp.h>

#include <net/checksum.h>

#include <rtnet.h>
#include <rtnet_internal.h>

struct rt_icmp_control
{
	void (*handler)(struct rtskb *skb);
	short	error;				/* This ICMP is classed as an error message */
};


/***
 *	icmp_sockets
 *	the registered sockets from any server
 */
struct rtsocket *icmp_sockets;          
//static unsigned short good_port=1024;


/***
 *	rt_icmp_discard - dummy function
 */
static void rt_icmp_discard(struct rtskb *skb)
{
}


/***
 *	rt_icmp_unreach - dummy function
 */
static void rt_icmp_unreach(struct rtskb *skb)
{
}



/***
 *	rt_icmp_redirect - dummy function
 */
static void rt_icmp_redirect(struct rtskb *skb)
{
}



/***
 *	rt_icmp_echo - 
 */
static void rt_icmp_echo(struct rtskb *skb)
{
}



/***
 *	rt_icmp_timestamp - 
 */
static void rt_icmp_timestamp(struct rtskb *skb)
{
}



/***
 *	rt_icmp_address - 
 */
static void rt_icmp_address(struct rtskb *skb)
{
}



/***
 *	rt_icmp_address_reply
 */
static void rt_icmp_address_reply(struct rtskb *skb)
{
}





struct rtsocket_ops rt_icmp_socket_ops = {
	bind:		NULL,
	connect:	NULL,
	listen:		NULL,
	accept:		NULL,
	recvmsg:	NULL,
	sendmsg:	NULL,
	close:		NULL,
#if 0        
	bind:		&rt_udp_bind,
	connect:	&rt_udp_connect,
	listen:		&rt_udp_listen,
	accept:		&rt_udp_accept,
	recvmsg:	&rt_udp_recvmsg,
	sendmsg:	&rt_udp_sendmsg,
	close:		&rt_udp_close,
#endif
};



/***
 *	rt_icmp_socket
 */
int rt_icmp_socket(struct rtsocket *sock) 
{
	sock->family=AF_INET;
	sock->typ=SOCK_DGRAM;
	sock->protocol=IPPROTO_ICMP;
	sock->ops = &rt_icmp_socket_ops;
	
	/* add to udp-socket-list */
	sock->next=icmp_sockets;
	if (icmp_sockets!=NULL)
		icmp_sockets->prev=sock;
	sock->prev=NULL;
	icmp_sockets=sock;
	
	return sock->fd;
}



/***
 *	rt_icmp_rcv
 */
int rt_icmp_rcv(struct rtskb *skb)
{
	return 0;
}



/***
 *	rt_icmp_rcv_err
 */
void rt_icmp_rcv_err (struct rtskb *skb) 
{
	rt_printk("RTnet: rt_icmp_rcv err\n");
}



static struct rt_icmp_control rt_icmp_pointers[NR_ICMP_TYPES+1] = {

	/* ECHO REPLY (0) */
	{ rt_icmp_discard,	0 },
	{ rt_icmp_discard,	1 },
	{ rt_icmp_discard,	1 },

	/* DEST UNREACH (3) */
	{ rt_icmp_unreach,	1 },

	/* SOURCE QUENCH (4) */
	{ rt_icmp_unreach,	1 },

	/* REDIRECT (5) */
	{ rt_icmp_redirect,	1 },
	{ rt_icmp_discard,	1 },
	{ rt_icmp_discard,	1 },

	/* ECHO (8) */
	{ rt_icmp_echo,	0 },
	{ rt_icmp_discard,	1 },
	{ rt_icmp_discard,	1 },

	/* TIME EXCEEDED (11) */
	{ rt_icmp_unreach,	1 },

	/* PARAMETER PROBLEM (12) */
	{ rt_icmp_unreach,	1 },

	/* TIMESTAMP (13) */
	{ rt_icmp_timestamp,	0  },

	/* TIMESTAMP REPLY (14) */
	{ rt_icmp_discard,	0 },

	/* INFO (15) */
	{ rt_icmp_discard,	0 },

	/* INFO REPLY (16) */
	{ rt_icmp_discard,	0 },

	/* ADDR MASK (17) */
	{ rt_icmp_address,	0  },

	/* ADDR MASK REPLY (18) */
	{ rt_icmp_address_reply, 0 }
};






/***
 *	ICMP-Initialisation	
 */
static struct rtinet_protocol icmp_protocol = {
	protocol:	IPPROTO_ICMP,
	handler:	&rt_icmp_rcv,
	err_handler:	&rt_icmp_rcv_err,
	socket:		&rt_icmp_socket,
};

/***
 *	rt_icmp_init
 */
void rt_icmp_init(void)
{
	icmp_sockets=NULL;
	rt_inet_add_protocol(&icmp_protocol);
}

/***
 *	rt_icmp_release
 */
void rt_icmp_release(void)
{
	rt_inet_del_protocol(&icmp_protocol);
	icmp_sockets=NULL;
}




